use std::any::TypeId;
use std::collections::HashMap;
use std::ops::Deref;
use std::rc::Rc;
use std::cell::RefCell;

pub(crate) mod event;

pub use event::ChannelSpec;

use crate::fleck_core;

pub type EventRoot = event::EventRoot<crate::Fleck>;

#[allow(unused_variables)]
pub trait Service: std::any::Any + crate::helper::AsAny {
    fn setup(&self, api: &crate::Fleck) {}
}

trait ServiceExt {
    fn build() -> Self;
}

impl<T: Service + Default> ServiceExt for T {
    fn build() -> Self {
        Default::default()
    }
}

#[derive(Default)]
pub struct ServiceStack {
    services: RefCell<Vec<Box<dyn Service>>>,

    service_map: RefCell<HashMap<TypeId, Box<dyn Service>>>,

    eroot: EventRoot,
    message_registry: fleck_core::msg::MessageRegistry,
}

impl ServiceStack {
    pub(crate) fn give_service<S: 'static + Default>(&self, api: &crate::Fleck)
    where
        Rc<S>: Service,
    {
        self.add_service(api, Rc::new(S::default()));
    }

    pub(crate) fn add_service<S: 'static>(&self, api: &crate::Fleck, srv: Rc<S>)
    where
        Rc<S>: Service,
    {
        srv.setup(api);

        self.service_map.borrow_mut()
            .insert(TypeId::of::<S>(), Box::new(srv.clone()));
        self.services.borrow_mut().push(Box::new(srv));
    }

    pub fn with_service<S: 'static, R, F: FnOnce(&S) -> R>(&self, f: F) -> R
    where
        Rc<S>: Service,
    {
        let id = TypeId::of::<S>();

        let svc = {
            let svcs = self.service_map.borrow();
            svcs.get(&id).expect("asked for service that doesn't exist!")
                .deref().as_any().downcast_ref::<Rc<S>>().unwrap().clone()
        };

        f(svc.as_ref())
    }

    pub fn event_root(&self) -> &EventRoot {
        &self.eroot
    }

    pub fn message_registry(&self) -> &fleck_core::msg::MessageRegistry {
        &self.message_registry
    }
}

/*
impl ServiceStack {
    pub(crate) fn process_incoming(&self, api: &crate::Fleck, packet: Packet) {
        self.eroot
            .channel::<fleck_core::ReceivePacketChannel>()
            .queue(packet);
        self.eroot.fire_all(api);
    }

    pub(crate) fn process_minor_tick(&self, api: &crate::Fleck) {
        self.eroot.channel::<fleck_core::MinorTickChannel>().queue(());
        self.eroot.fire_all(api);
    }

    pub(crate) fn process_major_tick(&self, api: &crate::Fleck) {
        self.eroot.channel::<fleck_core::MajorTickChannel>().queue(());
        self.eroot.fire_all(api);
    }

    pub(crate) fn create_io_channels(&mut self) {
        self.eroot.create_channel::<fleck_core::SendPacketChannel>();
        self.eroot.create_channel::<fleck_core::ReceivePacketChannel>();
        self.eroot.create_channel::<fleck_core::MinorTickChannel>();
        self.eroot.create_channel::<fleck_core::MajorTickChannel>();
    }
}
*/
